#include "ESPNow.hpp"
#include "LC709203F.h"
#include "NoDataAvailableException.hpp"
#include "dr26.hpp"
#include "f_deep_sleep.hpp"
#include <Arduino.h>
#include <SentecRainGaugeSensor.h>
#include <list>
#include <rs485.hpp>

// Execution time:
// FIXME: Boot: needs to be optimised. should be around 300 ms, can be <100ms
// https://github.com/makermoekoe/Picoclick-C3#speed-up-boot-process
// Part				[ms]
// Boot:			 300
// First setup: 	 150
// Data aquisition:  500	(4 Dendrometers)
// ESPNow setup: 	 220
// Sending 			 580
// TOTAL: 			1450

// Measurement Summary, 2 Dendrometers connected, measuring 4 Dendrometers:
// Duration:                    1.74 [s]
// Current consumption:       141.09 [mA*s]
// Average current:            81.16 [mA]

static const char *TAG = "MAIN";

LC709203F battery_monitor;

#define RS485Serial Serial2
#define RE_DE_PIN 19 // Line to pull high or low to receive or send data from RS485

#define RXPin 14 // Serial Receive pin
#define TXPin 15 // Serial Transmit pin

#define POWER_SWITCH_PIN_12V 12
#define POWER_SWITCH_PIN_5V 13

Forte_RS485 rs485;
SEM404 rainGaugeSensor{2, RE_DE_PIN};

ForteDR26 dr26_channel0;
ForteDR26 dr26_channel3;
ForteDR26 dr26_channel1;
ForteDR26 dr26_channel2;

// one loop takes ~2200 ms
void setup() {
    unsigned long ts = millis();
    Serial.begin(115200);
    rs485.setup();
    // Set the GPIO which conrtols the step up to OUTPUT
    gpio_set_direction(GPIO_NUM_32, GPIO_MODE_OUTPUT);

    // blinking led for debug
    // gpio_set_direction(GPIO_NUM_17, GPIO_MODE_OUTPUT);
    // gpio_set_level(GPIO_NUM_17, 1);

    battery_monitor.begin();

    DeepSleep::print_wakeup_reason();
    DeepSleep::bootCount++;
    ESP_LOGD(TAG, "Boot number: %d", DeepSleep::bootCount);
    // battery protection: go to deep sleep for unlimited time when voltage less than 3.2V
    if (battery_monitor.cellVoltage_mV() < 3200) {
        battery_monitor.setPowerMode(LC709203F_POWER_SLEEP);
        esp_deep_sleep_start();
    }

    gpio_set_level(GPIO_NUM_32, 1);
    // delay(100);
    dr26_channel1.setup();
    dr26_channel0.setChannel(0);
    dr26_channel1.setChannel(1);
    dr26_channel2.setChannel(2);
    dr26_channel3.setChannel(3);

    ESP_LOGD(TAG, "Setup took %ld ms", millis() - ts);

    try {
        // FIXME: put me into seperate trys? No data will be sent when 1 exception occurs

        ts = millis();
        //    auto messages0 = dr26_channel0.buildMessages();
        //    auto messages1 = dr26_channel1.buildMessages();
        //    auto messages2 = dr26_channel2.buildMessages();
        //    auto messages3 = dr26_channel3.buildMessages();
        //    auto messages4 = battery_monitor.buildMessages();
        Forte_RS485::powerOnRS485Sensors();
        auto soilMessage = rainGaugeSensor.buildMessages();
        Forte_RS485::powerOffRS485Sensors();
        // roughly takes 500ms, ~120ms for each adc channel, barely anything for battery monitor
        ESP_LOGD(TAG, "Reading data and building messages took %ld ms", millis() - ts);
        gpio_set_level(GPIO_NUM_32, 0);

        //    std::array<Message, 6> messages =
        //        {messages0.front(), messages1.front(), messages2.front(),
        //         messages3.front(), messages4.front(), soilMessage.front()};
        std::array<Message, 6> messages = {Message::nullMessage(), Message::nullMessage(), Message::nullMessage(),
                                           Message::nullMessage(), Message::nullMessage(), soilMessage.front()};

        ts = millis();
        espnow_setup();
        ESP_LOGD(TAG, "EPSNow setup took %ld ms", millis() - ts);

        if (messages.back().getClientDataPackage().getMeasurementData().getValue() > 20) {
            Forte_RS485::powerOnRS485Sensors();
            if (rainGaugeSensor.resetPrecipitation() == ErrorType::SEM404_COULD_NOT_RESET_PRECIPITATION) {
                ESP_LOGE(TAG, "Could not reset precipitation");
                auto precipitationErrorMessage = Message(
                    Measurement(ERROR_VALUE, rainGaugeSensor.address, NO_I2C_ADDRESS, MeasurementType::PRECIPITATION,
                                ErrorType::SEM404_COULD_NOT_RESET_PRECIPITATION),
                    rainGaugeSensor.getSensorInformation(), Time::getInstance().getEpochSeconds());
                std::array<Message, 6> messages_error = {precipitationErrorMessage, Message::nullMessage(),
                                                         Message::nullMessage(),    Message::nullMessage(),
                                                         Message::nullMessage(),    Message::nullMessage()};
                Message::sendMessages(messages_error);
            }
            Forte_RS485::powerOffRS485Sensors();
        }

        ts = millis();
        Message::sendMessages(messages);
        // roughly takes 3s in ideal conditions
        ESP_LOGD(TAG, "Sending messages took %ld ms", millis() - ts);

    } catch (const NoDataAvailableException &e) {
        std::cerr << e.what() << '\n';
    }
    // just to be safe in case exception happens above
    gpio_set_level(GPIO_NUM_32, 0);
    // keep it in deep sleep
    gpio_hold_en((gpio_num_t)GPIO_NUM_32);

    battery_monitor.setPowerMode(LC709203F_POWER_SLEEP);
    // battery protection: go to deep sleep for unlimited time when voltage less than 3.2V
    if (battery_monitor.cellVoltage_mV() > 3200) {
        DeepSleep::deep_sleep(2);
    } else {
        esp_deep_sleep_start();
    }
}

void loop() {}