//
// Created by zoe on 1/9/23.
//

#include "Utilities.h"

namespace SDUtilities {

    bool SDAvailable = true;

    void setSDAvailable(bool sdAvailable) {
        SDAvailable = sdAvailable;
    }

    // Used to try to remount SD card if it fails more than x times (100)
    int sdFailCounter = 0;

    bool isSDAvailable() {
        return SDAvailable;
    }

    void tryRemountingSD() {
        try {
            setupSDCard(SD_MISO, SD_MOSI, SD_SCLK, SD_CS);
            setSDAvailable(true);
        } catch (SDSetupException &e) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Couldn't remount SD card: %s\n", e.what());
        }
        sdFailCounter = 0;
    }

    bool checkSDAvailability(const String &errorMessage) {
        if (!isSDAvailable()) {
            sdFailCounter++;
            if (sdFailCounter % SD_CARD_FAIL_COUNTER == 0) {
                esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "SD card not available. Trying to remount...\n");
                tryRemountingSD();
            }
            String message = errorMessage + " (SD card not available, failed " + sdFailCounter + " times)\n";
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "%s", message.c_str());
            // if re-mount was successful, this will be true
            return isSDAvailable();
        }
        return true;
    };

    std::list<String> getFilesInDirectory(const String &dirname) {
        if (!checkSDAvailability("Couldn't get files in directory " + dirname)) {
            return {};
        }
        std::list<String> files;
        File dir = openDirectory(dirname);
        while (true) {
            File nextFile = dir.openNextFile();
            if (!nextFile) {
                break;
            }
            if (!nextFile.isDirectory()) {
                files.emplace_back(nextFile.name());
            }
            nextFile.close();
        }
        return files;
    }

    std::optional<String> getLastFileInDirectory(const String &dirname) {

        if (!checkSDAvailability("Couldn't get last file in directory " + dirname)) {
            return std::nullopt;
        }

        File root = openDirectory(dirname);
        root.rewindDirectory();

        File file = root.openNextFile();
        while (file) {
            File nextFile = root.openNextFile();
            if (!nextFile) {
                break;
            }
            file = nextFile;
        }
        // log
        if (file) {
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "Last file name: %s\n", file.name());
            return file.name();
        } else {
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "No file found\n");
            return std::nullopt;
        }
    }
    File openDirectory(const String &dirname) {

        if (!checkSDAvailability("Couldn't open directory " + dirname)) {
            throw;
        }

        File root = SD.open(dirname);
        if (!root) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Failed to open directory\n");
            throw;
        }
        if (!root.isDirectory()) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Not a directory\n");
            throw;
        }
        return root;
    }

    std::optional<String> getFirstFileNameInDirectory(const String &dirname) {

        if (!checkSDAvailability("Couldn't get first file in directory " + dirname)) {
            return std::nullopt;
        }

        File root = openDirectory(dirname);
        root.rewindDirectory();

        File file = root.openNextFile();
        if (file) {
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "file found: %s\n", file.name());
            return file.name();
        } else {
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "no file found\n");
            return std::nullopt;
        }
    }

    File openForWrite(const String &filePath) {

        if (!checkSDAvailability("Couldn't open file " + filePath + " for writing")) {
            throw;
        }

        File file = SD.open(filePath, FILE_WRITE);
        if (!file) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Failed to open file for writing\n");
            throw;
        }
        return file;
    }

    File openForRead(const String &filePath) {

        if (!checkSDAvailability("Couldn't open file " + filePath + " for reading")) {
            throw;
        }

        File file = SD.open(filePath, FILE_READ);
        if (!file) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Failed to open file for reading\n");
            throw;
        }
        return file;
    }

    void writeFile(const String &messageToBeSend, const String &filePath) {

        if (!checkSDAvailability("Couldn't write to file " + filePath)) {
            throw;
        }

        File file = openForWrite(filePath);
        if (file.print(messageToBeSend)) {
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "File written %s\n", filePath.c_str());
        } else {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Write failed %s\n", filePath.c_str());
            throw;
        }
        file.close();
    }

    String readFile(const String &filePath) {

        if (!checkSDAvailability("Couldn't read file " + filePath)) {
            throw;
        }

        File file = openForRead(filePath);

        String ret;

        while (file.available()) {
            ret += (char)file.read();
        }

        file.close();
        return ret;
    }

    void createDirectory(const String &dirname) {

        if (!checkSDAvailability("Couldn't create directory " + dirname)) {
            throw;
        }

        if (!SD.exists(dirname)) {
            SD.mkdir(dirname);
            esp_log_write(ESP_LOG_INFO, TAG_SDUTILITIES, "Created directory: %s\n", dirname.c_str());
        } else {
            esp_log_write(ESP_LOG_WARN, TAG_SDUTILITIES, "Directory already exists\n");
        }
    }

    void setupSDCard(int MISO, int MOSI, int SCLK, int CS) {
        SPI.begin(SCLK, MISO, MOSI, CS);
        if (!SD.begin(CS)) {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "Card MOUNT FAIL\n");
            throw SDSetupException("Card MOUNT FAIL");
        } else {
            uint32_t cardSize = SD.cardSize() / (1024 * 1024);
            String sdcardSizeString = "SDCard Size: " + String(cardSize) + "MB";
            esp_log_write(ESP_LOG_DEBUG, TAG_SDUTILITIES, "%s\n", sdcardSizeString.c_str());
        }
    }

    void saveStringToSDCard(const std::string &dataString) {

        if (!checkSDAvailability("Couldn't save string to SD card")) {
            throw SDCardException("Couldn't save string to SD card");
        }

        File dataFile = SD.open("/datalog.txt", FILE_APPEND);

        // if the file is available, write to it:
        if (dataFile) {
            if (dataString.length() > 0) {
                dataFile.println(dataString.c_str());
            }
            dataFile.close();
        }
        // if the file isn't open, pop up an error:
        else {
            esp_log_write(ESP_LOG_ERROR, TAG_SDUTILITIES, "error opening datalog.txt\n");
            throw SDCardException("error opening datalog.txt");
        }
    }
} // namespace SDUtilities

namespace ESPUtilities {

    String getMacAddressAsString(const uint8_t *mac) {
        String macAddress;
        for (int i = 0; i < 6; i++) {
            macAddress += String(mac[i], HEX);
        }
        esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "MAC: %s\n", macAddress.c_str());
        return macAddress;
    }

    void checkPeerExistence(const uint8_t *mac) {
        if (!esp_now_is_peer_exist(mac)) {
            esp_now_peer_info_t client = {};
            memcpy(client.peer_addr, mac, sizeof(uint8_t) * 6);
            client.encrypt = false;
            client.channel = 0;

            esp_err_t status = esp_now_add_peer(&client);
            if (status != ESP_OK) {
                esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Failed to add new Peer: %d", status);
            }
        }
    }

    void sendResponse(const uint8_t *mac) {
        response response = {};
        response.type = dataAck;
        esp_read_mac(response.mac, ESP_MAC_WIFI_STA);
        response.time = rtc.getEpoch();
        esp_err_t success = esp_now_send(mac, (uint8_t *)&response, sizeof(response));
    }
} // namespace ESPUtilities

// I don't think this does anything. Copied from the example
void turnOffLEDs() { // Set LED OFF
    pinMode(LED_PIN, OUTPUT);
    digitalWrite(LED_PIN, HIGH);

    pinMode(PWR_PIN, OUTPUT);
    digitalWrite(PWR_PIN, HIGH);
    delay(300);
    digitalWrite(PWR_PIN, LOW);
}

String dataPackageToLineProtocolString(const ClientDataPackage &dataPackage, const String &clientMacAddress) {
    String measurementType = dataPackage.getMeasurementData().getMeasurementTypeString().c_str();
    String hardwareName = dataPackage.getSensorInformation().getHardwareNameString().c_str();
    String timestamp = String(dataPackage.getTimestamp());
    String protocol = dataPackage.getSensorInformation().getProtocolString().c_str();
    String ic2address = String(dataPackage.getMeasurementData().getI2CAddress());
    String value = String(dataPackage.getMeasurementData().getValue());
    String channel = String(dataPackage.getMeasurementData().getChannel());

    String lineData = hardwareName + ",clientMac=" + clientMacAddress + ",protocol=" + protocol + ",channel=" + channel
                      + " " + measurementType + "=" + value + " " + timestamp;
    return lineData;
}

String documentToLineProtocolString(const DynamicJsonDocument &doc) {
    String measurementType = doc["measurementType"].as<String>();
    String sensorName = doc["hardwareName"].as<String>();
    String timestamp = doc["timestamp"].as<String>();
    String protocol = doc["protocolAddress"].as<String>();
    String ic2address = doc["ic2Address"].as<String>();
    String value = doc["value"].as<String>();
    String channel = doc["channel"].as<String>();
    String clientMac = doc["clientMac"].as<String>();

    String lineData = sensorName + ",clientMac=" + clientMac + ",protocol=" + protocol + ",channel=" + channel + " "
                      + measurementType + "=" + value + " " + timestamp;
    return lineData;
}
DynamicJsonDocument parseReceivedJsonData(char *data) {
    DynamicJsonDocument doc(250);
    auto error = deserializeJson(doc, data);
    if (error) {
        esp_log_write(ESP_LOG_ERROR, TAG_ESPNOW, "Error while parsing json: %s\n", error.c_str());
        // TODO error handling
    }
    return doc;
}

String documentToServerReadableString(const DynamicJsonDocument &doc) {
    StaticJsonDocument<1024> serverDoc;
    String hostMacAddressString = WiFi.macAddress();
    hostMacAddressString.replace(":", "");
    hostMacAddressString.toLowerCase();

    serverDoc["host"] = hostMacAddressString;
    serverDoc["client"] = doc["clientMac"].as<String>();
    serverDoc["sensorProtocol"] = doc["sensorProtocol"].as<String>();
    serverDoc["protocolAddress"] = doc["channel"].as<String>();
    serverDoc["i2cAddress"] = doc["i2cAddress"].as<String>();
    serverDoc["hardwareName"] = doc["hardwareName"].as<String>();
    // each value is a element in the readigs array
    JsonArray readings = serverDoc.createNestedArray("readings");
    JsonObject reading = readings.createNestedObject();
    reading["name"] = doc["measurementType"].as<String>();
    reading["value"] = doc["value"].as<double>();
    serverDoc["time"] = doc["timestamp"].as<uint32_t>();

    String serverString;
    serializeJson(serverDoc, serverString);
    esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Server readable data: %s\n", serverString.c_str());
    return serverString;
}

String dataPackageToServerReadableString(ClientDataPackage clientDataPackage, const String &clientMacAddress) {
    StaticJsonDocument<300> serverDoc;
    String hostMacAddressString = WiFi.macAddress();
    hostMacAddressString.replace(":", "");
    hostMacAddressString.toLowerCase();

    serverDoc["host"] = hostMacAddressString;
    serverDoc["client"] = clientMacAddress;
    serverDoc["sensorProtocol"] = clientDataPackage.getSensorInformation().getProtocolString();
    serverDoc["protocolAddress"] = clientDataPackage.getMeasurementData().getChannel();
    serverDoc["i2cAddress"] = clientDataPackage.getMeasurementData().getI2CAddress();
    serverDoc["hardwareName"] = clientDataPackage.getSensorInformation().getHardwareNameString();
    serverDoc["sensorError"] = clientDataPackage.getMeasurementData().getErrorTypeString();
    // each value is a element in the readigs array
    JsonArray readings = serverDoc.createNestedArray("readings");
    JsonObject reading = readings.createNestedObject();
    reading["name"] = clientDataPackage.getMeasurementData().getMeasurementTypeString();
    reading["value"] = clientDataPackage.getMeasurementData().getValue();
    serverDoc["time"] = clientDataPackage.getTimestamp();

    String serverString;
    serializeJson(serverDoc, serverString);
    esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Server readable data: %s\n", serverString.c_str());
    return serverString;
}